==> wimlib-0.4.3.tar.gz: wimlib-0.4.3/src/util.c <==
/*
 * util.c
 *
 * Some utility functions internal to wimlib, but not specific to the structures
 * declared in wimlib_internal.h.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2010 Carl Thijssen
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib_internal.h"
#include "endianness.h"
#include "sha1.h"


#include <iconv.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include <error.h>
#include <errno.h>
#include <stdarg.h>

/* True if WIMLIB is to print an informational message when an error occurs.
 * This can be turned off by calling wimlib_set_error_messages(false). */
static bool print_errors = true;

/* Print formatted error message to stderr. */
INTERNAL void wimlib_error(const char *format, ...)
{
	if (print_errors) {
		va_list va;
		va_start(va, format);
		fputs("WIMLIB: ", stderr);
		vfprintf(stderr, format, va);
		va_end(va);
	}
}

void wimlib_set_print_errors(bool show_error_messages)
{
	print_errors = show_error_messages;
}

INTERNAL void *xmalloc(u64 size)
{
	/* Watch for for casting u64 to size_t.  If size_t is only 32 bits, less
	 * memory will be allocated than we think. */
	if (sizeof(size_t) < sizeof(u64)) {
		if (size > 0xffffffffULL) {
			fprintf(stderr, "Cannot allocate %"PRIu64" bytes of memory "
					"when size_t is only %zu bytes!\n",
					size, sizeof(size_t));
			abort();
		}
	}
	void *p = malloc((size_t)size);
	if (p == NULL) {
		fprintf(stderr, "Out of memory! (Tried to allocate %lu bytes)\n", size);
		abort();
	}
	return p;
}

INTERNAL void *xzalloc(u64 size)
{
	void *p = xmalloc(size);
	return memset(p, 0, size);
}

INTERNAL void *xrealloc(void *ptr, u64 size)
{
	void *p = realloc(ptr, size);
	if (p == NULL) {
		fprintf(stderr, "Out of memory! (Tried tore allocate %lu bytes)\n", size);
		abort();
	}
	return p;
}


INTERNAL char *xstrdup(const char *s)
{
	if (s == NULL)
		return NULL;
	char *p = strdup(s);
	if (p == NULL) {
		fprintf(stderr, "strdup(): Out of memory!\n");
		abort();
	}
	return p;
}

static iconv_t cd_utf16_to_utf8 = (iconv_t)(-1);

INTERNAL char *utf16_to_utf8(const char *utf16_str, size_t utf16_len,
				size_t *utf8_len_ret)
{
	if (cd_utf16_to_utf8 == (iconv_t)(-1)) {
		cd_utf16_to_utf8 = iconv_open("UTF-8", "UTF-16LE");
		if (cd_utf16_to_utf8 == (iconv_t)-1) {
			wimlib_error("Failed to get conversion descriptor for "
					"converting UTF-16LE to UTF-8: %s\n", 
					strerror(errno));
			return NULL;
		}
	}
	size_t utf16_bytes_left  = utf16_len;

	/* XXX */
	size_t utf8_bytes_left   = utf16_len;

	char *utf8_str = xmalloc(utf8_bytes_left);
	char *orig_utf8_str = utf8_str;

	size_t num_chars_converted = iconv(cd_utf16_to_utf8, (char**)&utf16_str, 
			&utf16_bytes_left, &utf8_str, &utf8_bytes_left);

	if (num_chars_converted == (size_t)(-1)) {
		wimlib_error("Failed to convert UTF-16LE string to UTF-8 string: "
				"%s\n", strerror(errno));
		return NULL;
	}

	/* XXX */
	size_t utf8_len = utf16_len - utf8_bytes_left;

	*utf8_len_ret = utf8_len;
	orig_utf8_str[utf8_len] = '\0';
	return orig_utf8_str;
}

static iconv_t cd_utf8_to_utf16 = (iconv_t)(-1);

INTERNAL char *utf8_to_utf16(const char *utf8_str, size_t utf8_len, 
						size_t *utf16_len_ret)
{
	if (cd_utf8_to_utf16 == (iconv_t)(-1)) {
		cd_utf8_to_utf16 = iconv_open("UTF-16LE", "UTF-8");
		if (cd_utf8_to_utf16 == (iconv_t)-1) {
			wimlib_error("Failed to get conversion descriptor for "
					"converting UTF-8 to UTF-16LE: %s\n", 
					strerror(errno));
			return NULL;
		}
	}

	size_t utf8_bytes_left   = utf8_len;
	/* XXX */
	size_t utf16_capacity    = utf8_len * 4;
	size_t utf16_bytes_left  = utf16_capacity;

	char *utf16_str = xmalloc(utf16_capacity + 2);
	char *orig_utf16_str = utf16_str;


	size_t num_chars_converted = iconv(cd_utf8_to_utf16, (char**)&utf8_str, 
			&utf8_bytes_left, &utf16_str, &utf16_bytes_left);

	if (num_chars_converted == (size_t)(-1)) {
		wimlib_error("Failed to convert UTF-8 string to UTF-16LE string: "
				"%s\n", (errno == E2BIG) ? 
				"Not enough room in output buffer" : strerror(errno));
		return NULL;
	}

	/* XXX */
	size_t utf16_len = utf16_capacity - utf16_bytes_left;

	*utf16_len_ret = utf16_len;
	orig_utf16_str[utf16_len] = '\0';
	orig_utf16_str[utf16_len + 1] = '\0';
	return orig_utf16_str;
}

#define intervals_per_second (1000000000ULL / 100ULL)
#define days_per_year (365ULL)
#define seconds_per_day (3600ULL * 24ULL)
#define intervals_per_day (seconds_per_day * intervals_per_second)
#define intervals_per_year (intervals_per_day * days_per_year)
#define years_1601_to_1970 (1970ULL - 1601ULL)
#define leap_years_1601_to_1970 (years_1601_to_1970 / 4ULL - 3ULL)
#define intervals_1601_to_1970 (years_1601_to_1970 * intervals_per_year \
				+ leap_years_1601_to_1970 * intervals_per_day)

/* 
 * Returns the number of 100-nanosecond intervals that have elapsed since
 * 12:00 A.M., January 1, 1601 UTC.
 */
INTERNAL u64 get_timestamp()
{
	return (u64)intervals_1601_to_1970 + (u64)time(NULL) * intervals_per_second;
}

INTERNAL time_t ms_timestamp_to_unix(u64 timestamp)
{
	return (timestamp - intervals_1601_to_1970) / intervals_per_second;
}

static bool seeded = false;

INTERNAL void randomize_byte_array_with_alnum(u8 *p, size_t n)
{
	if (!seeded) {
		srand(time(NULL));
		seeded = true;
	}
	while (n--) {
		int r = rand() % 62;
		if (r < 26)
			*p++ = r + 'a';
		else if (r < 52)
			*p++ = r - 26 + 'A';
		else
			*p++ = r - 52 + '0';
	}
}

INTERNAL void randomize_byte_array(u8 *p, size_t n)
{
	if (!seeded) {
		srand(time(NULL));
		seeded = true;
	}
	while (n--)
		*p++ = rand();
}

/* Takes in a path of length @len in @buf, and transforms it into a string for
 * the path of its parent directory. */
INTERNAL void to_parent_name(char buf[], int len)
{
	int i = len - 1;
	while (i >= 0 && buf[i] == '/')
		i--;
	while (i >= 0 && buf[i] != '/')
		i--;
	while (i >= 0 && buf[i] == '/')
		i--;
	buf[i + 1] = '\0';
}

/* Like basename(), but does not modify path; it just returns a pointer to it.
 * */
INTERNAL const char *path_basename(const char *path)
{
	long i = strlen(path) - 1;
	while (i >= 0 && path[i] == '/')
		i--;
	while (i >= 0 && path[i] != '/')
		i--;
	return path + i + 1;
}

/* 
 * Splits a file path into the part before the first '/', or the entire name if
 * there is no '/', and the part after the first sequence of '/' characters.
 *
 * @path:  		The file path to split.
 * @first_part_len_ret: A pointer to an `int' into which the length of the
 * 				first part of the path will be returned.
 * @return:  		A pointer to the next part of the path, after the first
 * 				sequence of '/', or a pointer to the terminating 
 * 				null byte in the case of a path without any '/'.
 */
INTERNAL const char *path_next_part(const char *path, 
					int *first_part_len_ret)
{
	int i = 0;
	while (path[i] != '/' && path[i] != '\0')
		i++;
	*first_part_len_ret = i;
	const char *next_part = &path[i];
	while (*next_part == '/')
		next_part++;
	return next_part;
}

/* Returns the number of components of @path.  */
INTERNAL int get_num_path_components(const char *path)
{
	int num_components = 0;
	do {
		if (*path == '/') {
			num_components++;
			while (*++path == '/');
		}
	} while (*path++);
	return num_components;
}


/* 
 * Prints a string.  Printable characters are printed as-is, while unprintable
 * characters are printed as their octal escape codes. 
 *
 * @string: 	Pointer to the string.
 * @len: 	Length of the string. 
 */
INTERNAL void print_string(const u8 string[], u64 len)
{
	while (len--) {
		if (isprint(*string))
			putchar(*string++);
		else
			printf("\\%03hho", *string++);
	}
}

INTERNAL bool sha1sum(const char *filename, void *buf)
{
	FILE *fp = fopen(filename, "rb");
	if (fp == NULL) {
		wimlib_error("Cannot open the file `%s' for reading: %s\n", 
				filename, strerror(errno));
		return false;
	}
	int ret = sha1_stream(fp, buf);
	fclose(fp);
	return (ret == 0);
}

==> wimlib-0.4.5.tar.gz: wimlib-0.4.5/src/util.c <==
/*
 * util.c
 *
 * Some utility functions internal to wimlib, but not specific to the structures
 * declared in wimlib_internal.h.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2010 Carl Thijssen
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib_internal.h"
#include "endianness.h"
#include "sha1.h"


#include <iconv.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include <error.h>
#include <errno.h>
#include <stdarg.h>

/* True if WIMLIB is to print an informational message when an error occurs.
 * This can be turned off by calling wimlib_set_error_messages(false). */
static bool print_errors = true;

/* Print formatted error message to stderr. */
void wimlib_error(const char *format, ...)
{
	if (print_errors) {
		va_list va;
		va_start(va, format);
		fputs("WIMLIB ERROR: ", stderr);
		vfprintf(stderr, format, va);
		va_end(va);
	}
}

WIMLIBAPI void wimlib_set_print_errors(bool show_error_messages)
{
	print_errors = show_error_messages;
}

void *xmalloc(u64 size)
{
	/* Watch for for casting u64 to size_t.  If size_t is only 32 bits, less
	 * memory will be allocated than we think. */
	if (sizeof(size_t) < sizeof(u64)) {
		if (size > 0xffffffffULL) {
			fprintf(stderr, "Cannot allocate %"PRIu64" bytes of memory "
					"when size_t is only %zu bytes!\n",
					size, sizeof(size_t));
			abort();
		}
	}
	void *p = malloc((size_t)size);
	if (p == NULL) {
		fprintf(stderr, "Out of memory! (Tried to allocate %lu bytes)\n", size);
		abort();
	}
	return p;
}

void *xzalloc(u64 size)
{
	void *p = xmalloc(size);
	return memset(p, 0, size);
}

void *xrealloc(void *ptr, u64 size)
{
	void *p = realloc(ptr, size);
	if (p == NULL) {
		fprintf(stderr, "Out of memory! (Tried tore allocate %lu bytes)\n", size);
		abort();
	}
	return p;
}


char *xstrdup(const char *s)
{
	if (s == NULL)
		return NULL;
	char *p = strdup(s);
	if (p == NULL) {
		fprintf(stderr, "strdup(): Out of memory!\n");
		abort();
	}
	return p;
}

static iconv_t cd_utf16_to_utf8 = (iconv_t)(-1);


/* Converts a string in the UTF-16 encoding to a newly allocated string in the
 * UTF-8 encoding.  */
char *utf16_to_utf8(const char *utf16_str, size_t utf16_len,
				size_t *utf8_len_ret)
{
	if (cd_utf16_to_utf8 == (iconv_t)(-1)) {
		cd_utf16_to_utf8 = iconv_open("UTF-8", "UTF-16LE");
		if (cd_utf16_to_utf8 == (iconv_t)-1) {
			wimlib_error("Failed to get conversion descriptor for "
					"converting UTF-16LE to UTF-8: %s\n", 
					strerror(errno));
			return NULL;
		}
	}
	size_t utf16_bytes_left  = utf16_len;

	/* XXX */
	size_t utf8_bytes_left   = utf16_len;

	char *utf8_str = xmalloc(utf8_bytes_left);
	char *orig_utf8_str = utf8_str;

	size_t num_chars_converted = iconv(cd_utf16_to_utf8, (char**)&utf16_str, 
			&utf16_bytes_left, &utf8_str, &utf8_bytes_left);

	if (num_chars_converted == (size_t)(-1)) {
		wimlib_error("Failed to convert UTF-16LE string to UTF-8 string: "
				"%s\n", strerror(errno));
		return NULL;
	}

	/* XXX */
	size_t utf8_len = utf16_len - utf8_bytes_left;

	*utf8_len_ret = utf8_len;
	orig_utf8_str[utf8_len] = '\0';
	return orig_utf8_str;
}

static iconv_t cd_utf8_to_utf16 = (iconv_t)(-1);

/* Converts a string in the UTF-8 encoding to a newly allocated string in the
 * UTF-16 encoding.  */
char *utf8_to_utf16(const char *utf8_str, size_t utf8_len, 
						size_t *utf16_len_ret)
{
	if (cd_utf8_to_utf16 == (iconv_t)(-1)) {
		cd_utf8_to_utf16 = iconv_open("UTF-16LE", "UTF-8");
		if (cd_utf8_to_utf16 == (iconv_t)-1) {
			wimlib_error("Failed to get conversion descriptor for "
					"converting UTF-8 to UTF-16LE: %s\n", 
					strerror(errno));
			return NULL;
		}
	}

	size_t utf8_bytes_left   = utf8_len;
	/* XXX */
	size_t utf16_capacity    = utf8_len * 4;
	size_t utf16_bytes_left  = utf16_capacity;

	char *utf16_str = xmalloc(utf16_capacity + 2);
	char *orig_utf16_str = utf16_str;


	size_t num_chars_converted = iconv(cd_utf8_to_utf16, (char**)&utf8_str, 
			&utf8_bytes_left, &utf16_str, &utf16_bytes_left);

	if (num_chars_converted == (size_t)(-1)) {
		wimlib_error("Failed to convert UTF-8 string to UTF-16LE string: "
				"%s\n", (errno == E2BIG) ? 
				"Not enough room in output buffer" : strerror(errno));
		return NULL;
	}

	/* XXX */
	size_t utf16_len = utf16_capacity - utf16_bytes_left;

	*utf16_len_ret = utf16_len;
	orig_utf16_str[utf16_len] = '\0';
	orig_utf16_str[utf16_len + 1] = '\0';
	return orig_utf16_str;
}

#define intervals_per_second (1000000000ULL / 100ULL)
#define days_per_year (365ULL)
#define seconds_per_day (3600ULL * 24ULL)
#define intervals_per_day (seconds_per_day * intervals_per_second)
#define intervals_per_year (intervals_per_day * days_per_year)
#define years_1601_to_1970 (1970ULL - 1601ULL)
#define leap_years_1601_to_1970 (years_1601_to_1970 / 4ULL - 3ULL)
#define intervals_1601_to_1970 (years_1601_to_1970 * intervals_per_year \
				+ leap_years_1601_to_1970 * intervals_per_day)

/* 
 * Returns the number of 100-nanosecond intervals that have elapsed since
 * 12:00 A.M., January 1, 1601 UTC.
 */
u64 get_timestamp()
{
	return (u64)intervals_1601_to_1970 + (u64)time(NULL) * intervals_per_second;
}

/* Converts a timestamp as used in the WIM file to a UNIX timestamp as used in
 * the time() function. */
time_t ms_timestamp_to_unix(u64 timestamp)
{
	return (timestamp - intervals_1601_to_1970) / intervals_per_second;
}

static bool seeded = false;

/* Fills @n bytes pointed to by @p with random alphanumeric characters. */
void randomize_byte_array_with_alnum(u8 *p, size_t n)
{
	if (!seeded) {
		srand(time(NULL));
		seeded = true;
	}
	while (n--) {
		int r = rand() % 62;
		if (r < 26)
			*p++ = r + 'a';
		else if (r < 52)
			*p++ = r - 26 + 'A';
		else
			*p++ = r - 52 + '0';
	}
}

/* Fills @n bytes pointer to by @p with random numbers. */
void randomize_byte_array(u8 *p, size_t n)
{
	if (!seeded) {
		srand(time(NULL));
		seeded = true;
	}
	while (n--)
		*p++ = rand();
}

/* Takes in a path of length @len in @buf, and transforms it into a string for
 * the path of its parent directory. */
void to_parent_name(char buf[], int len)
{
	int i = len - 1;
	while (i >= 0 && buf[i] == '/')
		i--;
	while (i >= 0 && buf[i] != '/')
		i--;
	while (i >= 0 && buf[i] == '/')
		i--;
	buf[i + 1] = '\0';
}

/* Like the basename() function, but does not modify @path; it just returns a
 * pointer to it. */
const char *path_basename(const char *path)
{
	long i = strlen(path) - 1;
	while (i >= 0 && path[i] == '/')
		i--;
	while (i >= 0 && path[i] != '/')
		i--;
	return path + i + 1;
}

/* 
 * Splits a file path into the part before the first '/', or the entire name if
 * there is no '/', and the part after the first sequence of '/' characters.
 *
 * @path:  		The file path to split.
 * @first_part_len_ret: A pointer to an `int' into which the length of the
 * 				first part of the path will be returned.
 * @return:  		A pointer to the next part of the path, after the first
 * 				sequence of '/', or a pointer to the terminating 
 * 				null byte in the case of a path without any '/'.
 */
const char *path_next_part(const char *path, 
					int *first_part_len_ret)
{
	int i = 0;
	while (path[i] != '/' && path[i] != '\0')
		i++;
	*first_part_len_ret = i;
	const char *next_part = &path[i];
	while (*next_part == '/')
		next_part++;
	return next_part;
}

/* Returns the number of components of @path.  */
int get_num_path_components(const char *path)
{
	int num_components = 0;
	do {
		if (*path == '/') {
			num_components++;
			while (*++path == '/');
		}
	} while (*path++);
	return num_components;
}


/* 
 * Prints a string.  Printable characters are printed as-is, while unprintable
 * characters are printed as their octal escape codes. 
 *
 * @string: 	Pointer to the string.
 * @len: 	Length of the string. 
 */
void print_string(const u8 string[], u64 len)
{
	while (len--) {
		if (isprint(*string))
			putchar(*string++);
		else
			printf("\\%03hho", *string++);
	}
}

/* Calculates the SHA1 message digest given the name of a file.
 * @buf must point to a buffer of length 20 bytes into which the message digest
 * is written.
 *
 * Returns true on success, false on failure.
 */
bool sha1sum(const char *filename, void *buf)
{
	FILE *fp = fopen(filename, "rb");
	if (fp == NULL) {
		wimlib_error("Cannot open the file `%s' for reading: %s\n", 
				filename, strerror(errno));
		return false;
	}
	int ret = sha1_stream(fp, buf);
	fclose(fp);
	return (ret == 0);
}
